# -*- encoding: utf-8 -*-

import re, string, unicodedata, array

utf8TeX = {}

for withItem, texCode in [("DIAERESIS", "\\\"{%s}"),
                          ("ACUTE", "\\'{%s}"),
                          ("CARON", "\\v{%s}"),
                          ("CEDILLA", "\\c{%s}"),
                          ("CIRCUMFLEX", "\\^{%s}"),
                          ("GRAVE", "\\`{%s}"),
                          ("RING ABOVE", "\\r{%s}"),
                          ("TILDE", "\\~{%s}")]:
    for letter in string.ascii_uppercase:
        name = "LATIN CAPITAL LETTER %s WITH %s" % (letter, withItem)
        try:
            unicodeletter = unicodedata.lookup(name)
        except KeyError:
            pass
        else:
            utf8TeX[unicodeletter] = texCode % letter
        name = "LATIN SMALL LETTER %s WITH %s" % (letter, withItem)
        try:
            unicodeletter = unicodedata.lookup(name)
        except KeyError:
            pass
        else:
            utf8TeX[unicodeletter] = texCode % letter.lower()

greekLetters = [] # we want to keep their order
for greekLetterString, texCapital, texSmall in [("ALPHA", "A", "\\ensuremath{\\alpha}"),
                                                ("BETA", "B", "\\ensuremath{\\beta}"),
                                                ("GAMMA", "\\ensuremath{\\Gamma}", "\\ensuremath{\\gamma}"),
                                                ("DELTA", "\\ensuremath{\\Delta}", "\\ensuremath{\\delta}"),
                                                ("EPSILON", "E", "\\ensuremath{\\epsilon}"),
                                                ("ZETA", "Z", "\\ensuremath{\\zeta}"),
                                                ("ETA", "H", "\\ensuremath{\\eta}"),
                                                ("THETA", "\\ensuremath{\\Theta}", "\\ensuremath{\\theta}"),
                                                ("IOTA", "I", "\\ensuremath{\\iota}"),
                                                ("KAPPA", "K", "\\ensuremath{\\kappa}"),
                                                ("LAMDA", "\\ensuremath{\\Lambda}", "\\ensuremath{\\lambda}"),
                                                ("MU", "M", "\\ensuremath{\\mu}"),
                                                ("NU", "N", "\\ensuremath{\\nu}"),
                                                ("XI", "\\ensuremath{\\Xi}", "\\ensuremath{\\xi}"),
                                                ("OMICRON", "O", "o"),
                                                ("PI", "\\ensuremath{\\Pi}", "\\ensuremath{\\pi}"),
                                                ("RHO", "P", "\\ensuremath{\\rho}"),
                                                ("SIGMA", "\\ensuremath{\\Sigma}", "\\ensuremath{\\sigma}"),
                                                ("TAU", "T", "\\ensuremath{\\tau}"),
                                                ("UPSILON", "\\ensuremath{\\Upsilon}", "\\ensuremath{\\upsilon}"),
                                                ("PHI", "\\ensuremath{\\Phi}", "\\ensuremath{\\phi}"),
                                                ("CHI", "X", "\\ensuremath{\\chi}"),
                                                ("PSI", "\\ensuremath{\\Psi}", "\\ensuremath{\\psi}"),
                                                ("OMEGA", "\\ensuremath{\\Omega}", "\\ensuremath{\\omega}")]:
    capitalGreekLetter = unicodedata.lookup("GREEK CAPITAL LETTER %s" % greekLetterString)
    smallGreekLetter = unicodedata.lookup("GREEK SMALL LETTER %s" % greekLetterString)
    utf8TeX[capitalGreekLetter] = texCapital
    utf8TeX[smallGreekLetter] = texSmall
    greekLetters.append(capitalGreekLetter)
    greekLetters.append(smallGreekLetter)
greekLetters = "".join(greekLetters)

for unicodeletter, tex in [(u"Å", "\\AA{}"), (u"å", "\\aa{}"),
                           # regular LaTeX does not support a generic stroke technique
                           (unicodedata.lookup("LATIN CAPITAL LETTER L WITH STROKE"), "\L{}"),
                           (unicodedata.lookup("LATIN SMALL LETTER L WITH STROKE"), "\l{}"),
                           (u"Ø", "\O{}"), (u"ø", "\o{}"),
                           (u"ß", "\\ss{}"),
                           (u"°", "\\ensuremath{^\\circ}"),
                           (unicodedata.lookup("PLANCK CONSTANT OVER TWO PI"), "\\ensuremath{\\hbar}"),
                           (unicodedata.lookup("EURO SIGN"), "\\EUR{}")]:
    utf8TeX[unicodeletter] = tex

names = map(unicodedata.name, utf8TeX.keys())
names.sort()

printableASCIIChars = string.maketrans('', '')[32: 127]
sortableASCIIChars = "".join([char for char in printableASCIIChars
                              if char not in "!\"#$%&()*+,/:;<=>?@[\\]^_`{|}~"])
composedLatinLetters = "".join([unicodedata.lookup(name) for name in names
                                if name.startswith("LATIN CAPITAL LETTER") or name.startswith("LATIN SMALL LETTER")])
otherChars = "".join([unicodedata.lookup(name) for name in names
                      if unicodedata.lookup(name) not in composedLatinLetters + greekLetters])

allChars = printableASCIIChars + composedLatinLetters + greekLetters + otherChars

invalidTeXRepairers = [(re.compile(r"\\'([a-zA-Z])"), r"\\'{\1}"),
                       (re.compile(r'\\"([a-zA-Z])'), r'\\"{\1}'),
                       (re.compile(r'{\\ss}'), r'\ss{}')]
def translateInvalidTeX(s):
    if s.find("\\") != -1:
        for p, e in invalidTeXRepairers:
            s = re.sub(p, e, s)
    return s

def tex2utf8(s):
    for utf8, tex in utf8TeX.items():
        if not unicodedata.name(utf8).startswith("GREEK"):
            s = s.replace(tex, utf8)
    if s.find("\\") != -1:
        print s
    return s

def simplifyToGerman(s):
    s = s.replace("\\-", "").replace("\n", " ").replace("\r", " ")
    for c in composedLatinLetters:
        s = s.replace(utf8TeX[c], c)
    n = ""
    for c in s:
        #if c in sortableASCIIChars + u"äöüßÄÖÜ":
        if c in printableASCIIChars + u"äöüßÄÖÜ":
            n += c
        else:
            l = unicodedata.name(c).split()
            assert l[0] == "LATIN", s.encode("utf-8") + unicodedata.name(c)
            assert l[1] == "CAPITAL" or l[1] == "SMALL", s.encode("utf-8") + unicodedata.name(c)
            assert l[2] == "LETTER", s.encode("utf-8") + unicodedata.name(c)
            if l[1] == "SMALL":
                n += l[3].lower()
            else:
                n += l[3]
    #if n.find("\\") != -1:
    #    raise RuntimeError(n.encode("utf-8"))
    return n

def sortTest(org, useTeX):
    ur'''mark non-sortable parts of a string by an asterisk

    The string is sortable, when the function does not change it:

        >>> sortTest(u'Wobst', False)
        u'Wobst'

    However, in order to test whether a string is sortable, you should check whether the result contains an asterisk:

        >>> '*' not in sortTest(u'Wobst', False)
        True

    TeX expressions are disabled:

        >>> sortTest(ur"Andr\'{e}", False)
        u"Andr*'*e*"

    but can be turned on:

        >>> sortTest(ur"Andr\'{e}", True)
        u"Andr\\'{e}"

    Empty sort strings are disallowed:

        >>> sortTest(u'', False)
        u'*'
        >>> sortTest(u' ', False)
        u'*'
        >>> sortTest(u'.', False)
        u'*'
        >>> sortTest(u'-', False)
        u'*'

    Some further tests for certain code paths:

        >>> len(sortTest(u' Ä', False)) == 2
        True
        >>> '*' not in sortTest(u' Ö', False)
        True
        >>> '*' not in sortTest(u'-Ü', False)
        True
        >>> '*' not in sortTest(u'.ß', False)
        True
    '''
    if useTeX and org.find("\\") != -1:
        copy = org.replace("\\-", "  ")
        for char in composedLatinLetters:
            copy = copy.replace(utf8TeX[char], char*len(utf8TeX[char]))
    else:
        copy = org
    test = array.array('u')
    assert len(copy) == len(org)
    empty = True
    for copychar, orgchar in zip(copy, org):
        if copychar in sortableASCIIChars:
            if copychar in '0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ':
                empty = False
            test.append(orgchar)
        elif copychar in sortableASCIIChars + composedLatinLetters:
            empty = False
            test.append(orgchar)
        else:
            test.append(u"*")
    if empty:
        return u"*"*max(1, len(org))
    return test.tounicode()

def sortKey(org, useTeX, german=False):
    ur"""generate sort key for unicode string (with optional LaTeX expressions)

    Basically we replace everything by small ASCII characters:

        >>> sortKey(u'Wobst', False)
        u'wobst'

    Modified latin characters are 'downconverted':

        >>> sortKey(u'André', False)
        u'andre'

    However umlauts are converted depending on the setting of the german flag:

        >>> sortKey(u'Jörg', False)
        u'jorg'
        >>> sortKey(u'Jörg', False, german=True)
        u'joerg'

    Same for sharp s:

        >>> sortKey(u'abcßdef', False)
        u'abcsdef'
        >>> sortKey(u'abcßdef', False, german=True)
        u'abcssdef'

    LaTeX descriptions of composedLatinLetters are properly decoded, when useTeX is turned on:

        >>> sortKey(ur"Andr\'{e}", True)
        u'andre'

    However it comes to a failure when useTeX is disabled:

        >>> sortKey(ur"Andr\'{e}", False)
        Traceback (most recent call last):
            ...
        ValueError: invalid sort key for 'Andr\'{e}', character 'REVERSE SOLIDUS'

    LaTeX hyphen marks are supported:

        >>> sortKey(u'An\-dre', True)
        u'andre'

    but only when LaTeX is turned on:

        >>> sortKey(u'An\-dre', False)
        Traceback (most recent call last):
            ...
        ValueError: invalid sort key for 'An\-dre', character 'REVERSE SOLIDUS'

    Spaces and other characters are 'simplified':

        >>> sortKey(u'   a-b.c\'d  e  ', False)
        u'ab cd e'

    Empty sortKeys are not allowed:

        >>> sortKey(u'   -.\'  ', False)
        Traceback (most recent call last):
            ...
        ValueError: empty sort key for '   -.'  '

    Numbers are allowed and are kept:

        >>> sortKey(u' 1  234   ', False)
        u'1 234'

    Finally, invalid characters are of course rejected:

        >>> sortKey(u'abc*def', False)
        Traceback (most recent call last):
            ...
        ValueError: invalid sort key for 'abc*def', character 'ASTERISK'
    """
    if useTeX and org.find("\\") != -1:
        copy = org.replace("\\-", "").replace("-", "")
        for char in composedLatinLetters:
            copy = copy.replace(utf8TeX[char], char)
    else:
        copy = org.replace("-", "").replace("'", "")
    key = array.array('u')
    for char in copy:
        if char in sortableASCIIChars:
            if char in '0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ':
                key.append(char.lower())
            else:
                key.append(u' ')
        elif german and char in u"äÄ":
            key.append(u"a")
            key.append(u"e")
        elif german and char in u"öÖ":
            key.append(u"o")
            key.append(u"e")
        elif german and char in u"üÜ":
            key.append(u"u")
            key.append(u"e")
        elif char == u"ß":
            key.append(u"s")
            if german:
                key.append(u"s")
        else:
            l = unicodedata.name(char).split()
            if len(l) < 4 or l[0] != "LATIN" or (l[1] != "CAPITAL" and l[1] != "SMALL") or l[2] != "LETTER":
                raise ValueError("invalid sort key for '%s', character '%s'" % (org.encode("utf-8"), unicodedata.name(char)))
            key.append(l[3][0].lower().decode('ASCII'))
    key = key.tounicode()
    while key.find("  ") != -1:
        key = key.replace("  ", " ")
    key = key.strip()
    if not key:
        raise ValueError("empty sort key for '%s'" % org.encode("utf-8"))
    return key

def mixedCase(s, useTeX):
    """check whether we have a mixed char string

        >>> mixedCase(u"A", False)
        False
        >>> mixedCase(u"a", False)
        False
        >>> mixedCase(ur"\c{C}", True)
        False
        >>> mixedCase(ur"\c{C}", False)
        True
        >>> mixedCase(ur"\c{C}c", True)
        True
        >>> mixedCase(ur"\c{C}c A", True)
        False

    """
    small = capital = 0
    if useTeX and s.find("\\") != -1:
        for c in composedLatinLetters:
            s = s.replace(utf8TeX[c], c)
    for c in s:
        try:
            l = unicodedata.name(c).split()
        except ValueError:
            pass
        else:
            if l[0] == "LATIN" and l[2] == "LETTER":
                if l[1] == "CAPITAL":
                    capital += 1
                else:
                    small += 1
    return 0 < capital <= small

printableTeX = {"$": "{\\$}",
                "&": "{\\&}",
                "%": "{\\%}",
                "#": "{\\#}",
                "\"": "\mbox{''}", # \\textquotedbl is not available in OT1
                "<": "\mbox{\\textless}",
                ">": "\mbox{\\textgreater}",
                "_": "\mbox{\\textunderscore}",
                "^": "\mbox{\\textasciicircum}",
                "~": "\mbox{\\textasciitilde}",
                "{": "\mbox{\\textbraceleft}",
                "}": "\mbox{\\textbraceright}",
                "\\": "\mbox{\\textbackslash}",
                "|": "\mbox{\\textbar}"}

convertNothing = lambda x: unicode(x)
convertUTF8 = lambda x: "".join([utf8TeX.get(c, c) for c in unicode(x)])
convertUTF8NoTeX = lambda x: "".join([utf8TeX.get(c, printableTeX.get(c, c)) for c in unicode(x)])

if __name__ == "__main__":
    open("utf8.tex", "w").write(r"""\documentclass{article}
\usepackage{marvosym}
\begin{document}
%s
\end{document}
""" % ", ".join([printableTeX.has_key(char) and printableTeX[char] or char for char in printableASCIIChars] +
                [utf8TeX[char] for char in composedLatinLetters + greekLetters + otherChars]))
